// ViewWindow.cpp: implementation of the CViewWindow class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ViewWindow.h"
#include <math.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

#if 0
CViewWindow::CViewWindow(int *pDisplayPasses, double *pMaxJitter,
                         CImageView* pImage, CMouseUpdater* pMouseUpdater,
                         bool timeVarying)
    : m_pRenderPasses(pRenderPasses), m_pMaxJitter(pMaxJitter), m_image(pImage),
      m_mouseUpdater(pMouseUpdater), m_timeVarying(timeVarying)
{
    m_width = 0;
    m_height = 0;
    m_grabWindowX = 0;
    m_grabWindowY = 0;

    m_grabImageX = 0.0;
    m_grabImageY = 0.0;
    m_grabScale  = 0.0;
    m_uiMode = NoInput;
    m_timerRunning = false;
}
#endif

CViewWindow::CViewWindow(int *pRenderPasses, double *pMaxJitter,
                         CMouseUpdater* pMouseUpdater, bool timeVarying)
    : m_pRenderPasses(pRenderPasses), m_pMaxJitter(pMaxJitter),
      m_image(NULL), m_mouseUpdater(pMouseUpdater), m_timeVarying(timeVarying)
{
    m_width = 0;
    m_height = 0;
    m_grabWindowX = 0;
    m_grabWindowY = 0;
    m_mouseInputX = 0.0;
    m_mouseInputY = 0.0;

    m_grabImageX = 0.0;
    m_grabImageY = 0.0;
    m_grabScale  = 0.0;
    m_uiMode = NoInput;
    m_timerRunning = false;
}

CViewWindow::~CViewWindow()
{
    delete m_image;
    if (m_hWnd) ::DestroyWindow(m_hWnd);
}

void
CViewWindow::SetImageView(CImageView* pImage)
{
    m_image = pImage;
}

void
CViewWindow::CopyToClipboardResizing(int copyWidth, int copyPasses)
{
    // Simulate a resize&rezoom, re-render, CopyToClipboard, restore zoom&size.
    // Status: behaves identically to CopyToClipboard, i.e., doesn't resize.  
    // Maybe because I haven't provoked a repaint.
    if (m_image) {
        int tStartMS = timeGetTime();

        double extraZoom = (double) copyWidth / m_width;
        int copyHeight = (int) (m_height * extraZoom + 0.5);
        m_image->SetViewSize(copyWidth,copyHeight);
        double zoomWas = m_image->GetZoom();
        m_image->SetZoom(zoomWas/extraZoom);
        // Set up sufficiently large bitmap
        HDC hdc, hdcMem;
        HBITMAP hBitmap, hOld;
        hdc = ::GetDC(m_hWnd);
        hdcMem = ::CreateCompatibleDC(hdc);
        hBitmap = CreateCompatibleBitmap(hdc,copyWidth, copyHeight);
        hOld = (HBITMAP)SelectObject(hdcMem, hBitmap);

        // Render into the memory bitmap
        RECT rc;
        rc.left = 0; rc.right = copyWidth; rc.top = 0; rc.bottom = copyHeight;
        for (int pass=1 ; pass <= copyPasses;pass++) {
            m_image->Render(rc, copyWidth, copyHeight, pass, *m_pMaxJitter);
            m_image->Blit(m_hWnd, hdcMem, rc, copyWidth, copyHeight);
        }
        // BitBlt(hdcMem,0, 0,copyWidth, copyHeight, hdc, 0, 0, SRCCOPY);
        // Copy into clipboard
        ::OpenClipboard(m_hWnd);
        ::EmptyClipboard();
        ::SetClipboardData(CF_BITMAP, hBitmap);
        ::CloseClipboard();
        //SelectObject(hdcMem, hOld);

        // Restore the view transform.
        m_image->SetZoom(zoomWas);
        m_image->SetViewSize(m_width,m_height);
        // Provoke re-render, re-display
        Changed();
        if (timeGetTime() - tStartMS > 5000) {
            ::MessageBox(NULL, TEXT("Done"), TEXT("Pan: copy to clipboard"), MB_OK);
        }
    }
}

// Map pixPos back from view to image coords.  One dimension.
static inline double inverseView(int winSize, double zoom, int pan, int pixPos)
{
    return zoom * (pixPos  - winSize/2) + pan;
}


// Alter ycoordinates from positive-down to positive-up.  Also stash
// window width and height for later use.
static SIZE flipY(HWND hWnd, short* pWindowY)
{
    RECT rect; SIZE size;
    GetClientRect(hWnd, &rect);
    size.cx = rect.right  - rect.left;
    size.cy = rect.bottom - rect.top;

    *pWindowY = size.cy - *pWindowY - 1;

    return size;
}

LRESULT CViewWindow::OnClose(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    ::DestroyWindow(m_hWnd);
    return 0;
}

LRESULT CViewWindow::OnDestroy(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    // BAD idea to post a WM_QUIT here from within the bowels of
    // a control, since it will cause the message loop to give up the
    // ghost => app shutdown.
    //	PostQuitMessage(WM_QUIT);
    return TRUE;
}

LRESULT CViewWindow::OnCreate(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    return TRUE;
}

LRESULT CViewWindow::OnCommand(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    return TRUE;
}

LRESULT CViewWindow::OnSize(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    m_width  = LOWORD(lParam);
    m_height = HIWORD(lParam);
    if (m_image) {
        m_image->SetViewSize(m_width,m_height);
    }
    Changed(true);
    return TRUE;
}

LRESULT
CViewWindow::OnKeyDown(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    switch (wParam) {

    case VK_F5: {
        if (m_image) { // restart: reset viewing transform & rerender.
            m_image->SetZoom(1);
            m_image->SetTranslation(0,0);
            if (m_mouseUpdater) {
                m_mouseUpdater->UpdateMousePos(0,0);
            }
            m_image->Redisplay();
            Changed();
        }
        break;
    }
    }
    return 0;
}

LRESULT CViewWindow::OnChar(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{

    switch (wParam) {
    case 'r':  // restart: reset viewing transform & redisplay.
        OnKeyDown(uMsg, VK_F5,lParam,bHandled);
        break;

// Now done via "copy" and its Ctl-c accelerator, but not in standalone or
// viewer control.
//#if 0
    case '\x03':
        if (m_image) {
            CopyToClipboardResizing(1440,30);
        }
        break;
//#endif

    default:
        return FALSE;  // not handled
    }

    return TRUE; // msg handled.
}

// void CViewWindow::Redisplay() {}

#if 0
static BOOL MsgReady(HWND hWnd)
{
    MSG msg;
    return ::PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE);
}
#endif

LRESULT CViewWindow::OnTimer(UINT uMsg, WPARAM timerId, LPARAM lParam, BOOL& bHandled)
{
    if (timerId == displayTimer) {
        if (m_timeVarying)
            InvalidateRect(NULL,FALSE);
        else
            TriggerRenderPass();
    }
    else
        bHandled = FALSE;
    return 0;
}

// OnPaint can get called either because of an exposure event, a content
// change, finished render pass, or animation timer.
LRESULT
CViewWindow::OnPaint(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    PAINTSTRUCT ps;
    HDC hdc;
    if (m_image) {
        hdc = BeginPaint(&ps);
        if (m_timeVarying) {
            m_image->Render(ps.rcPaint, m_width, m_height,
                            m_nextPass, *m_pMaxJitter);
        }
        // Otherwise, just blit
        m_image->Blit(m_hWnd, hdc, ps.rcPaint, m_width, m_height);
        EndPaint(&ps);
    }
    return 0; // handled.
}

// Update moveX, moveY, and scale
LRESULT CViewWindow::OnMouseMove(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    short windowX, windowY;
    SIZE winSize; 

    windowX = LOWORD(lParam);
    windowY = HIWORD(lParam);

    winSize = flipY(m_hWnd, &windowY);
    
    if ((wParam & MK_RBUTTON) == MK_RBUTTON) {
        if (m_mouseUpdater) {
            m_mouseInputX = (float)m_image->GetZoom()*(windowX - winSize.cx/2);
            m_mouseInputY = (float)m_image->GetZoom()*(windowY - winSize.cy/2);
            m_mouseUpdater->UpdateMousePos(m_mouseInputX,m_mouseInputY);
        }
        return 0;
    }

    if (m_uiMode != NoInput) {
        // Calculate screen motion relative to mode start
        int pixDX = windowX - m_grabWindowX,
            pixDY = windowY - m_grabWindowY;
        if (pixDX != 0 || pixDY != 0) {
            if (m_uiMode == Panning) {
                // We want the grab point to track the mouse, so the
                // view transform we construct must map the mouse point to
                // the image-space grab point.  Thus:
                // 
                //    scale * (windowX - winSize.cx/2) + moveX == grabImageX
                //    scale * (windowY - winSize.cy/2) + moveY == grabImageY
                //    
                // Solving yields

                m_image->SetTranslation
                    ( m_grabImageX - m_image->GetZoom() * (windowX - winSize.cx/2)
                      , m_grabImageY - m_image->GetZoom() * (windowY - winSize.cy/2)
                        );
 
            } else {     // zooming
                // Use vertical motion to determine scale factor.  Window
                // height up doubles, down halves.  Remember though that
                // to stretch, we use a scale of < 1.
                m_image->SetZoom(m_grabScale * pow(2.0, - (double) pixDY / m_height));

                // We want to scale around the grab point.  In other
                // words, the viewing transform must continue to map the
                // grab screen point to the grab image point.  Thus
                // 
                //    scale * (grabWindowX - winSize.cx/2) + moveX == grabImageX
                //    scale * (grabWindowY - winSize.cy/2) + moveY == grabImageY
                //    
                // Solving yields

                m_image->SetTranslation( m_grabImageX - m_image->GetZoom() * (m_grabWindowX - winSize.cx/2)
                                         , m_grabImageY - m_image->GetZoom() * (m_grabWindowY - winSize.cy/2)
                    );
            } 

            // Trigger rerender.  If panning, we should really enable incremental repainting.
            Changed();
        }
    }
    return 0;
}


LRESULT CViewWindow::OnButton(UINT uMsg, WPARAM keys, LPARAM lParam, BOOL& bHandled)
{
    bool isDown = (uMsg == WM_LBUTTONDOWN || uMsg == WM_RBUTTONDOWN);
    bool isLeft = (uMsg == WM_LBUTTONDOWN || uMsg == WM_LBUTTONUP);
    short windowX, windowY;

    SIZE winSize;
    bool shift = (keys & MK_SHIFT) ? true : false;

    windowX = LOWORD(lParam);
    windowY = HIWORD(lParam);

    winSize = flipY(m_hWnd, &windowY);
    
    // grab mouse events (even if cursor goes outside window.)
    if (isDown) {
        SetCapture();
        if (isLeft) 
            SetFocus();
    }
    
    if (!isDown) {
        ReleaseCapture();
    }

    if (isDown) {
        // Start pan or zoom mode
        switch (m_uiMode) {
        case NoInput:
            // Start pan or zoom
            m_uiMode = shift ? Zooming : Panning;

            // Remember where we started, in screen coordinates ...
            m_grabWindowX = windowX; m_grabWindowY = windowY;
            // ... and in image coordinates:
            m_grabImageX = inverseView(winSize.cx, m_image->GetZoom(), (int)m_image->GetPanX(), m_grabWindowX);
            m_grabImageY = inverseView(winSize.cy, m_image->GetZoom(), (int)m_image->GetPanY(), m_grabWindowY);

            // Zoom will need the grab scale
            m_grabScale = m_image->GetZoom();
            break;
        default:
            break;
        }
    }
    else {  // release
        m_uiMode = NoInput;
    }

    // if the user hit the right mouse button, update the mouse position as
    // seen by the underlying image effect.
    if (!isLeft && m_mouseUpdater) {
        m_mouseUpdater->UpdateMousePos((float)m_image->GetZoom()*(windowX - winSize.cx/2),
                                       (float)m_image->GetZoom()*(windowY - winSize.cy/2));
    }

    return 0;
}

LRESULT CViewWindow::OnWheel(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    double imageX, imageY;
    short ctrl, zDelta;
    short cursorX, cursorY;
    SIZE winSize;

    // WM_MOUSEWHEEL coordinates are in *screen* coordinates,
    // and not in client coords like other mouse messgs.
    // (The reason for this is, I think, that WM_MOUSEWHEEL
    // is first delivered to focus window, with DefWindowProc()
    // propagating it upwards if unhandled. => you don't want
    // to adjust the coords at every HWND level.)
    POINT pt;
	    
    pt.x = LOWORD(lParam);
    pt.y = HIWORD(lParam);
    ScreenToClient(&pt);
    cursorX = (short)pt.x;
    cursorY = (short)pt.y;

    ctrl = LOWORD(wParam);
    zDelta = HIWORD(wParam);
    winSize = flipY(m_hWnd, &cursorY);

    // what's the image coordinates corresponding to the cursor's point?
    imageX = inverseView(winSize.cx, m_image->GetZoom(), (int)m_image->GetPanX(), cursorX);
    imageY = inverseView(winSize.cy, m_image->GetZoom(), (int)m_image->GetPanY(), cursorY);

    // multiply scale factor by (2^w) or 1/(2^w) (where 'w' is the number of
    // wheel ticks.)
    m_image->SetZoom(m_image->GetZoom()* pow (2, -((double) zDelta / WHEEL_DELTA)));

    // We want to scale around the current point.  In other
    // words, the image point corresponding to the current
    // cursor position must remain the same after scaling. So,
    // 
    //    scale * (cursorX - winSize.cx/2) + moveX == imageX
    //    scale * (cursorY - winSize.cy/2) + moveY == imageY
    //    
    // Solving yields

    m_image->SetTranslation(imageX - m_image->GetZoom() * (cursorX - winSize.cx/2),
                            imageY - m_image->GetZoom() * (cursorY - winSize.cy/2));


    Changed();
    return 0;
}
