//
// Classes implemented by a Pan image & used by the 
// viewer (and UI.)
// 
#ifndef __PANDISP_H__
#define __PANDISP_H__

//
// CLASS: PanImageDisplay
//
// This abstract class defines the methods used to 
// render a Pan image into a bitmap + a method which
// returns the title/window title for the filter/effect.
// 
class PanImageDisplay
{
 public:

  virtual void Display
            ( float t
#if 0
	    , float panX
	    , float panY
	    , float zoom
#endif
	    , int xStart
	    , int yStart
	    , int rWidth
	    , int rHeight
#if 0
	    , int width
	    , int height
#endif
	    , int outPad
	    , int output
	    ) =0;
  virtual void SetViewSize (int w, int h) =0;
  virtual void SetViewTransform (float pX, float pY, float zoom) =0;

  virtual char* WindowTitle() =0;
};

typedef struct UI UI;

class PanImagePhSDisplay
{
 public:

  virtual void Display
            ( int width
	    , int height
	    , int output
	    , int outPad
	    , int pProc
	    , int doProgress
	    ) =0;

  // Sigh, shouldn't have to lump this in here - revisit later.
  virtual UI* GetUI(void* pImg) =0;
  virtual void SetInputImage(int bits, int width, int height, int rowPad) =0;
  virtual void SetViewTransform(float pX, float pY, float z) =0;
  virtual void SetParamPtr(void* ptr) =0;
  virtual int  GetParamSize() =0;
  virtual char* WindowTitle() =0;
  virtual void Init() =0;
  virtual void SetViewSize (int w, int h) =0;

};

class PanImageUI
{
public:
  virtual UI* GetUI() =0;
};

//
// The UI notifies the underlying image of a value change
// of a particular parameter through the methods of 
// FloatUpdater and IntUpdater. 
//
class FloatUpdater
{
public:
  virtual float GetFloat(void* s) =0;
  virtual void  UpdateFloat(void* s, float newVal) =0;
};

class IntUpdater
{
public:
  virtual void UpdateInt(void* s, int newVal) =0;
  virtual int  GetInt(void* s) =0;
};

#endif
